#=========================================================================
# TestStructuralTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Provide a structural translator that fits testing purposes."""

from .. import StructuralTranslator


def mk_TestStructuralTranslator( _StructuralTranslator ):
  def make_indent( src, nindent ):
    """Add nindent indention to every line in src."""
    indent = '  '
    for idx, s in enumerate( src ):
      src[ idx ] = nindent * indent + s

  def get_string( obj ):
    """Return the string that identifies `obj`"""
    if isinstance(obj, type): return obj.__name__
    return str( obj )

  class TestStructuralTranslator( _StructuralTranslator ):
    """Testing translator that implements structural callback methods."""

    def rtlir_tr_vector_dtype( s, dtype ):
      return str( dtype )

    def rtlir_tr_struct_dtype( s, dtype ):
      return dtype.get_name()

    def rtlir_tr_unpacked_array_type( s, Type ):
      return "" if Type is None else repr( Type )

    def rtlir_tr_port_decls( s, port_decls ):
      decls = ''
      for decl in port_decls:
        if decl:
          make_indent( decl, 1 )
          decls += '\n' + '\n'.join( decl )
      return f'port_decls:{decls}\n'

    def rtlir_tr_port_decl( s, id_, Type, array_type, dtype ):
      if id_ not in ["clk", "reset"]:
        array_type = repr(Type) if not array_type else array_type
        return [f'port_decl: {id_} {array_type}']
      else:
        return ""

    def rtlir_tr_wire_decls( s, wire_decls ):
      decls = ''
      for decl in wire_decls:
        make_indent( decl, 1 )
        decls += '\n' + '\n'.join( decl )
      return f'wire_decls:{decls}\n'

    def rtlir_tr_wire_decl( s, id_, Type, array_type, dtype ):
      array_type = repr(Type) if not array_type else array_type
      return [f'wire_decl: {id_} {array_type}']

    def rtlir_tr_const_decls( s, const_decls ):
      decls = ''
      for decl in const_decls:
        if decl:
          make_indent( decl, 1 )
          decls += '\n' + '\n'.join( decl )
      return f'const_decls:{decls}\n'

    def rtlir_tr_const_decl( s, id_, Type, array_type, dtype, value ):
      array_type = repr(Type) if not array_type else array_type
      return [f'const_decl: {id_} {array_type}']

    def rtlir_tr_interface_port_decls( s, port_decls ):
      decls = [['interface_ports:']]
      for decl in port_decls:
        make_indent( decl, 1 )
        decls.append( [ decl[0] ] )
      return decls

    def rtlir_tr_interface_port_decl( s, m, id_, rtype, array_type ):
      rtype = repr(rtype) if not array_type else array_type
      return [f'interface_port: {id_} {rtype}']

    def rtlir_tr_interface_decls( s, ifc_decls ):
      decls = ''
      for decl in ifc_decls:
        if decl:
          make_indent( decl, 1 )
          decls += '\n' + '\n'.join( decl )
      return f'interface_decls:{decls}\n'

    def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, port_decls ):
      ifc_rtype = str(ifc_rtype) if not array_type else array_type
      ret = [f'interface_decl: {ifc_id} {ifc_rtype}']
      for decl in port_decls:
        make_indent( decl, 1 )
        ret.append( decl[0] )
      return ret

    def rtlir_tr_subcomp_port_decls( s, port_decls ):
      decls = [['component_ports:']]
      for decl in port_decls:
        if decl:
          make_indent( decl, 1 )
          decls.append( [ decl[0] ] )
      return decls

    def rtlir_tr_subcomp_port_decl( s, m, c_id, c_rtype, c_array_type, port_id,
        port_rtype, port_dtype, array_type ):
      if port_id not in ["clk", "reset"]:
        port_rtype = repr(port_rtype) if not array_type else array_type
        return [f'component_port: {port_id} {port_rtype}']
      else:
        return ""

    def rtlir_tr_subcomp_ifc_port_decls( s, ifc_port_decls ):
      decls = [['component_ifc_ports:']]
      for decl in ifc_port_decls:
        if decl:
          make_indent( decl, 1 )
          decls.append( [ decl[0] ] )
      return decls

    def rtlir_tr_subcomp_ifc_port_decl( s, m, c_id, c_rtype, c_array_type,
        ifc_id, ifc_rtype, ifc_array_type, port_id, port_rtype,
        port_array_type ):
      port_rtype = repr(port_rtype) if not port_array_type else port_array_type
      return [f'component_ifc_port: {port_id} {port_rtype}']

    def rtlir_tr_subcomp_ifc_decls( s, ifc_decls ):
      decls = [['component_ifcs:']]
      for ifc_decl in ifc_decls:
        for decl in ifc_decl:
          if decl:
            make_indent( decl, 1 )
            decls.append( [ decl[0] ] )
      return decls

    def rtlir_tr_subcomp_ifc_decl( s, m, c_id, c_rtype, c_array_type, ifc_id,
        ifc_rtype, ifc_array_type, ports ):
      ifc_rtype = repr(ifc_rtype) if not ifc_array_type else ifc_array_type
      decls = [[f'component_ifc: {ifc_id} {ifc_rtype}']]
      for decl in ports:
        if decl:
          make_indent( decl, 1 )
          decls.append( [ decl[0] ] )
      return decls

    def rtlir_tr_subcomp_decls( s, subcomps ):
      decls = ''
      for decl in subcomps:
        make_indent( decl, 1 )
        decls += '\n' + '\n'.join( decl )
      return f'component_decls:{decls}\n'

    def rtlir_tr_subcomp_decl( s, m, c_id, c_rtype, c_array_type, port_conns, ifc_conns ):
      c_rtype = repr(c_rtype) if not c_array_type else c_array_type
      ret = [f'component_decl: {c_id} {c_rtype}']
      for port in port_conns:
        make_indent( port, 1 )
        ret.append( port[0] )
      for ifc in ifc_conns:
        make_indent( ifc, 1 )
        ret.append( ifc[0] )
      return ret

    def rtlir_tr_connections( s, connections ):
      conns = ''
      for conn in connections:
        if conn:
          make_indent( conn, 1 )
          conns += '\n' + '\n'.join( conn )
      return f'connections:{conns}\n'

    def rtlir_tr_connection( s, wr, rd ):
      if "clk" not in wr and "reset" not in wr:
        return [f'connection: {wr} -> {rd}']

    def rtlir_tr_bit_selection( s, base_signal, index ):
      return f'BitSel {base_signal} {index}'

    def rtlir_tr_part_selection( s, base_signal, start, stop ):
      return f'PartSel {base_signal} {start} {stop}'

    def rtlir_tr_port_array_index( s, base_signal, index ):
      return f'PortArrayIdx {base_signal} {index}'

    def rtlir_tr_wire_array_index( s, base_signal, index ):
      return f'WireArrayIdx {base_signal} {index}'

    def rtlir_tr_const_array_index( s, base_signal, index ):
      return f'ConstArrayIdx {base_signal} {index}'

    def rtlir_tr_packed_index( s, base_signal, index ):
      return f'PackedIndex {base_signal} {index}'

    def rtlir_tr_interface_array_index( s, base_signal, index ):
      return f'IfcArrayIdx {base_signal} {index}'

    def rtlir_tr_component_array_index( s, base_signal, index ):
      return f'CompArrayIdx {base_signal} {index}'

    def rtlir_tr_struct_attr( s, base_signal, attr ):
      return f'StructAttr {base_signal} {attr}'

    def rtlir_tr_interface_attr( s, base_signal, attr ):
      return f'IfcAttr {base_signal} {attr}'

    def rtlir_tr_subcomp_attr( s, base_signal, attr ):
      return f'SubCompAttr {base_signal} {attr}'

    def rtlir_tr_current_comp_attr( s, base_signal, attr ):
      return f'CurCompAttr {attr}'

    def rtlir_tr_current_comp( s, comp_id, comp_rtype ):
      return ''

    def rtlir_tr_var_id( s, var_id ):
      return var_id

    def rtlir_tr_literal_number( s, nbits, value ):
      return f'Bits{nbits}({int(value)})'

    def rtlir_tr_component_unique_name( s, c_rtype ):
      comp_name = c_rtype.get_name()
      comp_params = c_rtype.get_params()
      assert comp_name
      for arg_name, arg_value in comp_params:
        assert arg_name != ''
        comp_name += '__' + arg_name + '_' + get_string(arg_value)
      return comp_name

  return TestStructuralTranslator

TestStructuralTranslator = mk_TestStructuralTranslator( StructuralTranslator )



#=========================================================================
# DelayPipeCL.py
#=========================================================================
# This delay pipe models a inelasic pipeline queue with enq/deq interfaces
#
# Author : Shunning Jiang
# Date   : May 7, 2018

from collections import deque
from copy import deepcopy

from pymtl3 import *

# This delay pipe is for cycle-level performance modeling purpose

class DelayPipeDeqCL( Component ):

  @non_blocking( lambda s: s.pipeline[0] is None )
  def enq( s, msg ):
    assert s.pipeline[0] is None
    s.pipeline[0] = deepcopy(msg)

  @non_blocking( lambda s: s.pipeline[-1] is not None )
  def deq( s ):
    ret = s.pipeline[-1]
    s.pipeline[-1] = None
    return ret

  @non_blocking( lambda s: True )
  def peek( s ):
    assert s.pipeline[-1] is not None
    return s.pipeline[-1]

  def construct( s, delay=5, trace_len=0 ):

    s.delay = delay

    s.trace_len = trace_len

    if delay == 0: # This is essentially a bypass queue
      s.pipeline = [ None ]

      s.add_constraints(
        M(s.enq) < M(s.deq),  # bypass behavior
      )

    else: # delay >= 1, pipe behavior
      s.pipeline = deque( [None]*(delay+1), maxlen=(delay+1) )

      @s.update
      def up_delay():
        if s.pipeline[-1] is None:
          s.pipeline.rotate()

      # Model decoupled pipe behavior to cut cyclic dependencies.
      # Basically no matter in what order s.deq and s.enq are called,
      # the outcomes are the same as long as up_delay is called before
      # both of them.

      # NOTE THAT this up_delay affects ready signal, we need to mark it
      # before enq.rdy
      s.add_constraints(
        U(up_delay) < M(s.peek),
        U(up_delay) < M(s.deq),
        U(up_delay) < M(s.deq.rdy),
        U(up_delay) < M(s.enq),
        U(up_delay) < M(s.enq.rdy),
      )

  def line_trace( s ):
    return "[{}]".format( "".join( [ " " if x is None else "*" for x in list(s.pipeline)[:-1] ] ) )

class DelayPipeSendCL( Component ):

  def enq_pipe( s, msg ):
    assert s.pipeline[0] is None
    s.pipeline[0] = deepcopy(msg)

  def enq_rdy_pipe( s ):
    return s.pipeline[0] is None

  def construct( s, delay=5 ):

    s.send = CallerIfcCL()

    s.delay = delay

    if delay == 0: # combinational behavior
      s.enq = CalleeIfcCL()
      connect( s.enq, s.send )

    else: # delay >= 1, pipe behavior
      s.enq = CalleeIfcCL( Type=None, method=s.enq_pipe, rdy=s.enq_rdy_pipe )
      s.pipeline = deque( [None]*delay, maxlen=delay )

      @s.update
      def up_delay():
        if s.pipeline[-1] is not None:
          if s.send.rdy():
            s.send( s.pipeline[-1] )
            s.pipeline[-1] = None
            s.pipeline.rotate()
        else:
          s.pipeline.rotate()

      s.add_constraints(
        M(s.enq) > U(up_delay),  # pipe behavior
        M(s.enq.rdy) > U(up_delay),  # pipe behavior
      )

  def line_trace( s ):
    if s.delay > 0:
      return "[{}]".format( "".join( [ " " if x is None else "*" for x in s.pipeline ] ) )
    return ""



#=========================================================================
# test_utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 21, 2019
"""Test utilities used by RTLIR tests."""

from contextlib import contextmanager

import pytest


@pytest.fixture
def do_test( request ):
  """Call `local_do_test` of the requesting module."""
  return request.module.local_do_test

@contextmanager
def expected_failure( exception = Exception, msg = None ):
  """Mark one test case as should-fail.

  Not to be confused with pytest.xfail, which is commonly used to mark
  tests related to unimplemented functionality. This test only passes when
  it throws an expected exception.
  """
  try:
    yield
  except exception as e:
    if msg is None or e.args[0].find( msg ) != -1:
      return
    else:
      raise
  raise Exception( 'expected-to-fail test unexpectedly passed!' )

def get_parameter( name, func ):
  """Return the parameter for `name` arg of `func`"""
  try:
    for mark in func.pytestmark:
      if mark.name == 'parametrize':
        # Find the position of the given name
        pos = -1
        for i, arg in enumerate( mark.args[0].split() ):
          if arg == name:
            pos = i
            break
        if pos == -1:
          raise Exception( f'{func} does not have parameter named {name}!' )
        if len(mark.args[0].split()) == 1:
          return mark.args[1]
        return list(map(lambda x: x[pos], mark.args[1]))
  except AttributeError:
    raise Exception( f'given function {func} does not have pytest marks!' )



"""
========================================================================
TestMemory
========================================================================
A behavioral Test Memory which is parameterized based on the number of
memory request/response ports. This version is a little different from
the one in pclib because we actually use the memory messages correctly
in the interface.

Author : Shunning Jiang
Date   : Mar 12, 2018
"""

from pymtl3 import *
from pymtl3.stdlib.fl import MemoryFL
from pymtl3.stdlib.ifcs import MemMsgType, mk_mem_msg
from pymtl3.stdlib.ifcs.mem_ifcs import MemMinionIfcCL

from .DelayPipeCL import DelayPipeDeqCL, DelayPipeSendCL

# BRGTC2 custom MemMsg modified for RISC-V 32

#- - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - -
#-------------------------------------------------------------------------
# BRGTC2
#-------------------------------------------------------------------------
# The AMO implementations (and MemMsg) has been updated to match RISC-V.
#
# There is also a small fix to the AMO ops to handle signed ops. The AMO
# operations act on the bitwidth of the processor architecture, so the
# read_data from the TestMemory used with AMOs cannot just be the memory
# request message size (e.g., 128b):
#
#         read_data = Bits( s.data_nbits )
#
# It must instead be the number of bytes matching the bitwidth in the
# processor (e.g., 32b):
#
#         read_data = Bits( nbytes*8 )
#
# Otherwise for example we would be reading 128b from the memory and
# comparing that to the 32b value from the request message.
#
#-------------------------------------------------------------------------
#- - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - -

class MemoryCL( Component ):

  # Magical methods

  def read_mem( s, addr, size ):
    return s.mem.read_mem( addr, size )

  def write_mem( s, addr, data ):
    return s.mem.write_mem( addr, data )

  # Actual stuff
  def construct( s, nports, mem_ifc_dtypes=[mk_mem_msg(8,32,32), mk_mem_msg(8,32,32)], latency=1, mem_nbytes=2**20 ):

    # Local constants

    s.nports = nports
    req_classes  = [ x for (x,y) in mem_ifc_dtypes ]
    resp_classes = [ y for (x,y) in mem_ifc_dtypes ]

    s.mem = MemoryFL( mem_nbytes )

    # Interface

    s.ifc = [ MemMinionIfcCL( req_classes[i], resp_classes[i] ) for i in range(nports) ]

    # Queues
    req_latency = min(1, latency)
    resp_latency = latency - req_latency

    s.req_qs  = [ DelayPipeDeqCL( req_latency )( enq = s.ifc[i].req ) for i in range(nports) ]
    s.resp_qs = [ DelayPipeSendCL( resp_latency )( send = s.ifc[i].resp ) for i in range(nports) ]

    @s.update
    def up_mem():

      for i in range(s.nports):

        if s.req_qs[i].deq.rdy() and s.resp_qs[i].enq.rdy():

          # Dequeue memory request message

          req = s.req_qs[i].deq()
          len_ = int(req.len)
          if len_ == 0: len_ = req_classes[i].data_nbits >> 3

          if   req.type_ == MemMsgType.READ:
            resp = resp_classes[i]( req.type_, req.opaque, 0, req.len,
                                    s.mem.read( req.addr, len_ ) )

          elif req.type_ == MemMsgType.WRITE:
            s.mem.write( req.addr, len_, req.data )
            # FIXME do we really set len=0 in response when doing subword wr?
            # resp = resp_classes[i]( req.type_, req.opaque, 0, req.len, 0 )
            resp = resp_classes[i]( req.type_, req.opaque, 0, 0, 0 )

          else: # AMOS
            resp = resp_classes[i]( req.type_, req.opaque, 0, req.len,
               s.mem.amo( req.type_, req.addr, len_, req.data ) )

          s.resp_qs[i].enq( resp )

  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------
  # TODO: better line trace.

  def line_trace( s ):
    return "|".join( [ x[0].line_trace() + x[1].line_trace() for x in zip(s.req_qs, s.resp_qs) ] )



"""
========================================================================
Placeholder_test.py
========================================================================

Author : Shunning Jiang
Date   : June 1, 2019
"""
from collections import deque

from pymtl3.datatypes import *
from pymtl3.dsl import (
    WR,
    CalleeIfcCL,
    CalleePort,
    CallerPort,
    Component,
    InPort,
    M,
    OutPort,
    Placeholder,
    U,
    Wire,
    connect,
    method_port,
    non_blocking,
)
from pymtl3.dsl.errors import InvalidPlaceholderError, LeftoverPlaceholderError

from .sim_utils import simple_sim_pass


def test_placeholder_no_upblk():

  class Wrong( Placeholder, Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      @s.update
      def up_x():
        s.out = s.in_ + 1
  try:
    a = Wrong()
    a.elaborate()
  except InvalidPlaceholderError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidPlaceholderError.")

def test_placeholder_no_constraints():

  class Wrong( Placeholder, Component ):
    def construct( s, nbits=0 ):
      s.in_ = InPort ( mk_bits(nbits) )
      s.out = OutPort( mk_bits(nbits) )
      s.add_constraints( U(s.out) < WR(s.out) ) # this is obvious wrong

  try:
    a = Wrong()
    a.elaborate()
  except InvalidPlaceholderError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidPlaceholderError.")

class Foo( Placeholder, Component ):
  def construct( s, nbits=0 ):
    s.in_ = InPort ( mk_bits(nbits) )
    s.out = OutPort( mk_bits(nbits) )

    # Nothing here

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class SomeMsg:
  def __init__( s, a=0, b=0 ):
    s.a = Bits16(a)
    s.b = Bits32(b)

class FooStruct( Placeholder, Component ):
  def construct( s, nbits=0 ):
    s.in_ = InPort ( mk_bits(nbits) )
    s.out = OutPort( SomeMsg )

    # Nothing here

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class Real( Component ):
  def construct( s, nbits=0 ):
    s.in_ = InPort ( mk_bits(nbits) )
    s.out = OutPort( mk_bits(nbits) )
    @s.update
    def up_x2():
      s.out = s.in_ << 1

  def line_trace( s ):
    return "{}>{}".format( s.in_, s.out )

class Foo_wrap( Component ):
  def construct( s, nbits=0 ):
    s.in_ = InPort ( mk_bits(nbits) )
    s.out = OutPort( mk_bits(nbits) )
    s.w   = Wire( mk_bits(nbits) )
    connect( s.w, s.out )

    s.inner = Foo( 32 )( in_ = s.in_, out = s.w )

  def line_trace( s ):
    return s.inner.line_trace()

def test_dumb_foo():
  foo = Foo( 32 )
  foo.elaborate()
  # should elaborate just fine ..

def test_dumb_foo_cannot_simulate():
  foo = Foo( 32 )
  foo.elaborate()
  # should elaborate just fine ..
  try:
    simple_sim_pass( foo )
  except LeftoverPlaceholderError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown LeftoverPlaceholderError.")

def test_foo_as_writer():
  foo_wrap = Foo_wrap( 32 )

  foo_wrap.elaborate()
  try:
    simple_sim_pass( foo_wrap )
  except LeftoverPlaceholderError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown LeftoverPlaceholderError.")

def test_foo_field_as_writer():

  class FooStruct_wrap( Component ):
    def construct( s ):
      s.in_ = InPort ( Bits16 )
      s.out = OutPort( Bits32 )

      s.inner = FooStruct( 16 )( in_ = s.in_ )
      connect( s.inner.out.b, s.out )

    def line_trace( s ):
      return s.inner.line_trace()

  foo_wrap = FooStruct_wrap()
  foo_wrap.elaborate()
  try:
    simple_sim_pass( foo_wrap )
  except LeftoverPlaceholderError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown LeftoverPlaceholderError.")

def test_foo_replaced_by_real():
  foo_wrap = Foo_wrap( 32 )

  foo_wrap.elaborate()
  foo_wrap.replace_component( foo_wrap.inner, Real, check=True )

  simple_sim_pass( foo_wrap )
  foo_wrap.sim_reset()

  foo_wrap.in_ = Bits32(100)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 200

  foo_wrap.in_ = Bits32(3)
  foo_wrap.tick()
  print(foo_wrap.line_trace())
  assert foo_wrap.out == 6


def test_cl_methodport_placeholder():

  class FooCL( Placeholder, Component ):
    def construct( s ):
      s.enq = CalleePort()
      s.deq = CalleePort()

  class FooCL_wrap( Component ):
    def construct( s ):
      s.enq = CalleePort()
      s.deq = CalleePort()
      s.foo = FooCL()( enq = s.enq, deq = s.deq )

  class FooCL_top( Component ):
    def construct( s ):
      s.x = FooCL_wrap()

      s.counter = 0
      @s.update
      def up_enq():
        s.x.enq(s.counter + 1)
        s.counter += 1

      @s.update
      def up_recv():
        print(s.x.deq())

  class RealQueue( Component ):
    @method_port
    def enq( s, msg ):
      s.q.appendleft( msg )
    @method_port
    def deq( s ):
      return s.q.pop()

    def construct( s ):
      s.q = deque()
      s.add_constraints( M(s.enq) < M(s.deq) )

  a = FooCL_top()
  a.elaborate()

  a.replace_component( a.x.foo, RealQueue )

  simple_sim_pass( a )
  a.tick()
  a.tick()
  a.tick()

def test_cl_interface_placeholder():

  class FooCL( Placeholder, Component ):
    def construct( s, queue_type="none" ):
      s.enq = CalleeIfcCL()
      s.deq = CalleeIfcCL()

  class FooCL_wrap( Component ):
    def construct( s ):
      s.enq = CalleeIfcCL()
      s.deq = CalleeIfcCL()
      s.foo = FooCL()( enq = s.enq, deq = s.deq )

  class FooCL_top( Component ):
    def construct( s ):
      s.x = FooCL_wrap()

      s.received = None
      s.counter = 0
      @s.update
      def up_enq():
        if s.x.enq.rdy():
          s.x.enq(s.counter + 1)
          s.counter += 1
        else:
          print("enq not rdy")

      @s.update
      def up_recv():
        if s.x.deq.rdy():
          s.received = s.x.deq()
          print(s.received)
        else:
          print("deq not rdy")

  class RealQueue( Component ):
    def construct( s, queue_type ):
      s.element = None

      if queue_type == "bypass":
        s.add_constraints( M( s.enq ) < M( s.deq ) )
      else:
        assert queue_type == "pipe"
        s.add_constraints( M( s.deq ) < M( s.enq ) )

    def empty( s ):
      return s.element is None

    @non_blocking( lambda s : s.empty() )
    def enq( s, ele ):
      s.element = ele

    @non_blocking( lambda s: not s.empty() )
    def deq( s ):
      ret = s.element
      s.element = None
      return ret

  a = FooCL_top()
  b = FooCL_top()

  a.set_param( "top.x.foo.construct", queue_type="pipe" )
  b.set_param( "top.x.foo.construct", queue_type="bypass" )

  a.elaborate()
  b.elaborate()

  a.replace_component( a.x.foo, RealQueue )
  b.replace_component( b.x.foo, RealQueue )

  simple_sim_pass( a )
  simple_sim_pass( b )

  a.tick()
  b.tick()
  assert a.received == None # pipe queue cannot receive the packet in the first cycle
  assert b.received == 1 # bypass queue receive the packet in the first cycle
  a.tick()
  b.tick()
  assert a.received == 1 # pipe queue cannot receive the packet in the first cycle
  assert b.received == 2 # bypass queue receive the packet in the first cycle

  a.tick()
  b.tick()
  assert a.received == 2 # pipe queue cannot receive the packet in the first cycle
  assert b.received == 3 # bypass queue receive the packet in the first cycle



"""
========================================================================
Test sinks
========================================================================
Test sinks with CL and RTL interfaces.

Author : Yanghui Ou
  Date : Mar 11, 2019
"""

from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, RecvRTL2SendCL, enrdy_to_str

#-------------------------------------------------------------------------
# TestSinkCL
#-------------------------------------------------------------------------

class TestSinkCL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0,
                 arrival_time=None, cmp_fn=lambda a, b : a == b ):

    s.recv.Type = Type

    # [msgs] and [arrival_time] must have the same length.
    if arrival_time is not None:
      assert len( msgs ) == len( arrival_time )

    s.idx          = 0
    s.cycle_count  = 0
    s.msgs         = list( msgs )
    s.arrival_time = None if not arrival_time else list( arrival_time )
    s.cmp_fn       = cmp_fn
    s.error_msg    = ''

    s.all_msg_recved = False
    s.done_flag      = False

    s.count = initial_delay
    s.intv  = interval_delay

    s.recv_called = False

    @s.update
    def up_sink_count():
      # Raise exception at the start of next cycle so that the errored
      # line trace gets printed out
      if s.error_msg:
        raise Exception( s.error_msg )

      # Tick one more cycle after all message is received so that the
      # exception gets thrown
      if s.all_msg_recved:
        s.done_flag = True

      if s.idx >= len( s.msgs ):
        s.all_msg_recved = True

      if not s.reset:
        s.cycle_count += 1
      else:
        s.cycle_count = 0

      # if recv was called in previous cycle
      if s.recv_called:
        s.count = s.intv
      elif s.count != 0:
        s.count -= 1
      else:
        s.count = 0

      s.recv_called = False

    s.add_constraints(
      U( up_sink_count ) < M( s.recv ),
      U( up_sink_count ) < M( s.recv.rdy )
    )

  @non_blocking( lambda s: s.count==0 )
  def recv( s, msg ):
    assert s.count == 0, "Invalid en/rdy transaction! Sink is stalled (not ready), but receives a message."

    # Sanity check
    if s.idx >= len( s.msgs ):
      s.error_msg = ( 'Test Sink received more msgs than expected!\n'
                      f'Received : {msg}' )

    # Check correctness first
    elif not s.cmp_fn( msg, s.msgs[ s.idx ] ):
      s.error_msg = (
        f'Test sink {s} received WRONG message!\n'
        f'Expected : { s.msgs[ s.idx ] }\n'
        f'Received : { msg }'
      )

    # Check timing if performance regeression is turned on
    elif s.arrival_time and s.cycle_count > s.arrival_time[ s.idx ]:
      s.error_msg = (
        f'Test sink {s} received message LATER than expected!\n'
        f'Expected msg : {s.msgs[ s.idx ]}\n'
        f'Expected at  : {s.arrival_time[ s.idx ]}\n'
        f'Received msg : {msg}\n'
        f'Received at  : {s.cycle_count}'
      )

    else:
      s.idx += 1
      s.recv_called = True

  def done( s ):
    return s.done_flag

  # Line trace
  def line_trace( s ):
    return "{}".format( s.recv )

#-------------------------------------------------------------------------
# TestSinkRTL
#-------------------------------------------------------------------------

class TestSinkRTL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0,
                 arrival_time=None, cmp_fn=lambda a, b : a == b ):

    # Interface

    s.recv = RecvIfcRTL( Type )

    # Components

    s.sink    = TestSinkCL( Type, msgs, initial_delay, interval_delay,
                            arrival_time, cmp_fn )
    s.adapter = RecvRTL2SendCL( Type )

    connect( s.recv,         s.adapter.recv )
    connect( s.adapter.send, s.sink.recv    )

  def done( s ):
    return s.sink.done()

  # Line trace

  def line_trace( s ):
    return "{}".format( s.recv )



#=========================================================================
# BehavioralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L3 behavioral RTLIR generation pass."""

from pymtl3.datatypes import Bits, is_bitstruct_class
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRGenL2Pass import BehavioralRTLIRGeneratorL2


class BehavioralRTLIRGenL3Pass( BasePass ):
  def __call__( s, m ):
    """ generate RTLIR for all upblks of m """
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()
    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL3( m )

    upblks = {
      'CombUpblk' : get_ordered_upblks(m),
      'SeqUpblk'  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        upblk_info = m.get_update_block_info( blk )
        upblk = visitor.enter( blk, upblk_info[-1] )
        upblk.is_lambda = upblk_info[0]
        upblk.src       = upblk_info[1]
        upblk.lino      = upblk_info[2]
        upblk.filename  = upblk_info[3]
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = upblk

class BehavioralRTLIRGeneratorL3( BehavioralRTLIRGeneratorL2 ):
  def __init__( s, component ):
    super().__init__( component )

  def visit_Call( s, node ):
    """Return behavioral RTLIR of a method call.

    At L3 we need to support the syntax of struct instantiation in upblks.
    This is achieved by function calls like `struct( 1, 2, 0 )`.
    """
    obj = s.get_call_obj( node )
    if is_bitstruct_class( obj ):
      if len(node.args) < 1:
        raise PyMTLSyntaxError( s.blk, node,
          'at least one value should be provided to struct instantiation!' )
      values = [s.visit(arg) for arg in node.args]
      ret = bir.StructInst( obj, values )
      ret.ast = node
      return ret

    else:
      return super().visit_Call( node )



#=========================================================================
# SVStructuralTranslatorL2.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL2 import (
    StructuralTranslatorL2,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt

from ...util.utility import make_indent
from .SVStructuralTranslatorL1 import SVStructuralTranslatorL1


class SVStructuralTranslatorL2(
    SVStructuralTranslatorL1, StructuralTranslatorL2 ):

  #-----------------------------------------------------------------------
  # Data types
  #-----------------------------------------------------------------------

  def rtlir_tr_packed_array_dtype( s, dtype ):
    sub_dtype = dtype.get_sub_dtype()
    if isinstance( sub_dtype, rdt.Vector ):
      sub_dtype_template = s.rtlir_tr_vector_dtype( sub_dtype )
    elif isinstance( sub_dtype, rdt.Struct ):
      sub_dtype_template = s.rtlir_tr_struct_dtype( sub_dtype )
    else:
      raise Exception(f"unsupported data type {sub_dtype} in packed array!")
    dim_str = "".join( f"[{size-1}:0]" for size in dtype.get_dim_sizes() )
    str_list = sub_dtype_template['decl'].split()
    if '[' in str_list[-2]:
      str_list[-2] = dim_str + str_list[-2]
    else:
      str_list = str_list[:-1] + [ dim_str ] + [ str_list[-1] ]
    return {
      'def' : '',
      'const_decl' : ' '.join( str_list ),
      'decl' : ' '.join( str_list ),
      'ndim' : dtype.get_dim_sizes(),
      'raw_dtype' : dtype
    }

  def rtlir_tr_struct_dtype( s, dtype ):
    dtype_name = dtype.get_name()
    field_decls = []

    for id_, _dtype in dtype.get_all_properties().items():

      if isinstance( _dtype, rdt.Vector ):
        decl = s.rtlir_tr_vector_dtype( _dtype )['decl'].format(**locals())
      elif isinstance( _dtype, rdt.PackedArray ):
        decl = s.rtlir_tr_packed_array_dtype( _dtype )['decl'].format(**locals())
      elif isinstance( _dtype, rdt.Struct ):
        decl = s.rtlir_tr_struct_dtype( _dtype )['decl'].format(**locals())
      else:
        assert False, \
          f'unrecoganized field type {_dtype} of struct {dtype_name}!'
      field_decls.append( decl + ';' )

    make_indent( field_decls, 1 )
    field_decl = '\n'.join( field_decls )

    file_info = dtype.get_file_info()

    return {
      'def' : \
f"""\
typedef struct packed {{
{field_decl}
}} {dtype_name};
""",
      'const_decl' : f'{dtype_name} {{id_}}',
      'decl' : f'{dtype_name} {{id_}}',
      'raw_dtype' : dtype
    }

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def gen_array_param( s, n_dim, dtype, array ):
    if not n_dim:
      if isinstance( dtype, rdt.Struct ):
        return s.rtlir_tr_struct_instance( dtype, array )
      else:
        return super().gen_array_param( n_dim, dtype, array )
    else:
      return super().gen_array_param( n_dim, dtype, array )

  #-----------------------------------------------------------------------
  # Signal oeprations
  #-----------------------------------------------------------------------

  def rtlir_tr_packed_index( s, base_signal, index ):
    return f'{base_signal}[{index}]'

  def rtlir_tr_struct_attr( s, base_signal, attr ):
    return f'{base_signal}.{attr}'

  def rtlir_tr_struct_instance( s, dtype, struct ):
    def _gen_packed_array( dtype, n_dim, array ):
      if not n_dim:
        if isinstance( dtype, rdt.Vector ):
          return s.rtlir_tr_literal_number( dtype.nbits, array )
        elif isinstance( dtype, rdt.Struct ):
          return s.rtlir_tr_struct_instance( dtype, array )
        else:
          raise Exception(f"unrecognized data type {dtype}!")
      else:
        ret = []
        for i in reversed( range( n_dim[0]) ):
          ret.append( _gen_packed_array( dtype, n_dim[1:], array[i] ) )
        if n_dim[0] > 1:
          cat_str = "{" + ", ".join( ret ) + "}"
        else:
          cat_str = ", ".join( ret )
        return f"{{ {cat_str} }}"
    ret = []
    for name, Type in dtype.get_all_properties().items():
      field = getattr( struct, name )
      if isinstance( Type, rdt.Vector ):
        _ret = s.rtlir_tr_literal_number( Type.nbits, field )
      elif isinstance( Type, rdt.Struct ):
        _ret = s.rtlir_tr_struct_instance( Type, field )
      elif isinstance( Type, rdt.PackedArray ):
        n_dim = Type.get_dim_sizes()
        sub_dtype = Type.get_sub_dtype()
        _ret = _gen_packed_array( sub_dtype, n_dim, field )
      else:
        assert False, f"unrecognized data type {Type}!"
      ret.append( _ret )
    cat_str = ", ".join( ret )
    return f"{{ {cat_str} }}"



#=========================================================================
# TranslationImport_adhoc_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 15, 2019
"""Test ad-hoc components with yosys-SystemVerilog translation and import."""

import pytest

from pymtl3.passes.rtlir.util.test_utility import get_parameter
from pymtl3.stdlib.test import TestVectorSimulator

from ..translation.behavioral.test.YosysBehavioralTranslatorL1_test import (
    test_yosys_behavioral_L1,
)
from ..translation.behavioral.test.YosysBehavioralTranslatorL2_test import (
    test_yosys_behavioral_L2,
)
from ..translation.behavioral.test.YosysBehavioralTranslatorL3_test import (
    test_yosys_behavioral_L3,
)
from ..translation.behavioral.test.YosysBehavioralTranslatorL4_test import (
    test_yosys_behavioral_L4,
)
from ..translation.behavioral.test.YosysBehavioralTranslatorL5_test import (
    test_yosys_behavioral_L5,
)
from ..translation.structural.test.YosysStructuralTranslatorL1_test import (
    test_yosys_structural_L1,
)
from ..translation.structural.test.YosysStructuralTranslatorL2_test import (
    test_yosys_structural_L2,
)
from ..translation.structural.test.YosysStructuralTranslatorL3_test import (
    test_yosys_structural_L3,
)
from ..translation.structural.test.YosysStructuralTranslatorL4_test import (
    test_yosys_structural_L4,
)
from ..TranslationImportPass import TranslationImportPass

XFAILED_TESTS = [
    # incoherent translation result of Yosys backend
    # the translated design is correct after minor transformation
    # in logic synthesis, but verilator simulation is incorrect
    'CaseConnectLiteralStructComp',
]

def run_test( case ):
  try:
    _m = case.DUT()
    _m.elaborate()
    _m.yosys_translate_import = True
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, case.TEST_VECTOR, case.TV_IN, case.TV_OUT )
    sim.run_test()
  finally:
    try:
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

@pytest.mark.parametrize(
  'case', list(filter(lambda x: x.__name__ not in XFAILED_TESTS,
          get_parameter('case', test_yosys_behavioral_L1) + \
          get_parameter('case', test_yosys_behavioral_L2) + \
          get_parameter('case', test_yosys_behavioral_L3) + \
          get_parameter('case', test_yosys_behavioral_L4) + \
          get_parameter('case', test_yosys_behavioral_L5) + \
          get_parameter('case', test_yosys_structural_L1) + \
          get_parameter('case', test_yosys_structural_L2) + \
          get_parameter('case', test_yosys_structural_L3) + \
          get_parameter('case', test_yosys_structural_L4) ) )
)
def test_yosys_translation_import_adhoc( case ):
  run_test( case )



"""
========================================================================
src_sink_test
========================================================================
Tests for test sources and test sinks.

Author : Yanghui Ou
  Date : Mar 11, 2019
"""
import pytest

from pymtl3 import *

from .test_sinks import TestSinkCL, TestSinkRTL
from .test_srcs import TestSrcCL, TestSrcRTL

#-------------------------------------------------------------------------
# TestHarnessSimple
#-------------------------------------------------------------------------
# Test a single pair of test src/sink.

class TestHarnessSimple( Component ):

  def construct( s, MsgType, SrcType, SinkType, src_msgs, sink_msgs ):

    s.src  = SrcType ( MsgType, src_msgs  )
    s.sink = SinkType( MsgType, sink_msgs )

    connect( s.src.send, s.sink.recv  )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} > {}".format( s.src.line_trace(), s.sink.line_trace() )

  def run_sim( s, max_cycles=100 ):
    s.apply( SimulationPass() )
    ncycles = 0
    print("")
    print("{:2}:{}".format( ncycles, s.line_trace() ))
    while not s.done() and ncycles < max_cycles:
      s.tick()
      ncycles += 1
      print("{:2}:{}".format( ncycles, s.line_trace() ))

    # Check timeout
    assert ncycles < max_cycles

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

def test_cl_no_delay():
  msgs  = [ Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]
  th = TestHarnessSimple( Bits16, TestSrcCL, TestSinkCL, msgs, msgs )
  th.run_sim()

# int_msgs = [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ]
bit_msgs = [ Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ),
             Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ),
             Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]

arrival0 = [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 ]
arrival1 = [ 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21 ]
arrival2 = [ 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33 ]
arrival3 = [ 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33 ]
arrival4 = [ 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60 ]

@pytest.mark.parametrize(
  ('Type', 'msgs', 'src_init',  'src_intv',
   'sink_init', 'sink_intv', 'arrival_time' ),
  [
    ( Bits16, bit_msgs,  0,  0, 0, 0, arrival0 ),
    # ( int,    int_msgs, 10,  0, 0, 0, arrival1 ),
    ( Bits16, bit_msgs, 10,  1, 0, 0, arrival2 ),
    ( Bits16, bit_msgs, 10,  0, 0, 1, arrival3 ),
    ( Bits16, bit_msgs,  3,  4, 5, 3, arrival4 )
  ]
)
def test_src_sink_cl( Type, msgs, src_init,  src_intv,
                      sink_init, sink_intv, arrival_time ):
  th = TestHarnessSimple( Type, TestSrcCL, TestSinkCL, msgs, msgs )
  th.set_param( "top.src.construct",
    initial_delay  = src_init,
    interval_delay = src_intv,
  )
  th.set_param( "top.sink.construct",
    initial_delay  = sink_init,
    interval_delay = sink_intv,
    arrival_time   = arrival_time,
  )
  th.run_sim()

@pytest.mark.parametrize(
  ('Type', 'msgs', 'src_init',  'src_intv',
   'sink_init', 'sink_intv', 'arrival_time' ),
  [
    ( Bits16, bit_msgs,  0,  0, 0, 0, arrival0 ),
    # ( int,    int_msgs, 10,  0, 0, 0, arrival1 ),
    ( Bits16, bit_msgs, 10,  1, 0, 0, arrival2 ),
    ( Bits16, bit_msgs, 10,  0, 0, 1, arrival3 ),
    ( Bits16, bit_msgs,  3,  4, 5, 3, arrival4 )
  ]
)
def test_src_sink_rtl( Type, msgs, src_init,  src_intv,
                       sink_init, sink_intv, arrival_time ):
  th = TestHarnessSimple( Type, TestSrcRTL, TestSinkRTL, msgs, msgs )
  th.set_param( "top.src.construct",
    initial_delay  = src_init,
    interval_delay = src_intv,
  )
  th.set_param( "top.sink.construct",
    initial_delay  = sink_init,
    interval_delay = sink_intv,
    arrival_time   = arrival_time,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# Adaptive composition test
#-------------------------------------------------------------------------
# This test attempts to mix-and-match different levels of test srcs and
# sinks for all possible combinations -- cl/cl, rtl/cl, cl/rtl, rtl/rtl.
# It also creates multiple src/sink pairs to stress the management of
# multiple instances of the same adapter class

class TestHarness( Component ):

  def construct( s, src_level, sink_level, MsgType, src_msgs, sink_msgs,
                 src_init, src_intv,
                 sink_init, sink_interval, arrival_time=None ):
    s.num_pairs = 2

    if src_level == 'cl':
      s.srcs = [ TestSrcCL ( MsgType, src_msgs, src_init, src_intv )
                  for i in range(s.num_pairs) ]
    elif src_level == 'rtl':
      s.srcs = [ TestSrcRTL( MsgType, src_msgs, src_init, src_intv )
                  for i in range(s.num_pairs) ]
    else:
      raise

    if sink_level == 'cl':
      s.sinks = [ TestSinkCL( MsgType, sink_msgs, sink_init, sink_interval, arrival_time )
                  for i in range(s.num_pairs) ]
    elif sink_level == 'rtl':
      s.sinks = [ TestSinkRTL( MsgType, sink_msgs, sink_init, sink_interval, arrival_time )
                  for i in range(s.num_pairs) ]
    else:
      raise

    # Connections
    for i in range(s.num_pairs):
      connect( s.srcs[i].send, s.sinks[i].recv )

  def done( s ):
    for i in range(s.num_pairs):
      if not s.srcs[i].done() or not s.sinks[i].done():
        return False
    return True

  def line_trace( s ):
    return "{} >>> {}".format( "|".join( [ x.line_trace() for x in s.srcs ] ),
                               "|".join( [ x.line_trace() for x in s.sinks ] ) )

  def run_sim( s, max_cycles=100 ):
    s.apply( SimulationPass() )
    ncycles = 0
    print("")
    print("{:2}:{}".format( ncycles, s.line_trace() ))
    while not s.done() and ncycles < max_cycles:
      s.tick()
      ncycles += 1
      print("{:2}:{}".format( ncycles, s.line_trace() ))

    # Check timeout
    assert ncycles < max_cycles

test_case_table = []
for src in ['cl', 'rtl']:
  for sink in ['cl', 'rtl']:
    test_case_table += [
      ( src, sink, bit_msgs,  0,  0, 0, 0, arrival0 ),
      # ( src, sink, int_msgs, 10,  0, 0, 0, arrival1 ),
      ( src, sink, bit_msgs, 10,  1, 0, 0, arrival2 ),
      ( src, sink, bit_msgs, 10,  0, 0, 1, arrival3 ),
      ( src, sink, bit_msgs,  3,  4, 5, 3, arrival4 ),
    ]

@pytest.mark.parametrize(
  ('src_level', 'sink_level', 'msgs',
   'src_init',  'src_intv', 'sink_init', 'sink_intv', 'arrival_time' ),
  test_case_table,
)
def test_adaptive( src_level, sink_level, msgs, src_init,  src_intv,
                        sink_init, sink_intv, arrival_time ):
  th = TestHarness( src_level, sink_level, Bits16, msgs, msgs,
                    src_init,  src_intv, sink_init,
                    sink_intv, arrival_time )
  th.run_sim()

#-------------------------------------------------------------------------
# Error message test
#-------------------------------------------------------------------------

def test_error_more_msg():
  try:
    th = TestHarnessSimple(
      Bits16, TestSrcCL, TestSinkCL,
      src_msgs  = [ b16(0xface), b16(0xface) ],
      sink_msgs = [ b16(0xface) ],
    )
    th.run_sim()
  except Exception as e:
    return
  raise Exception( 'Failed to detect error!' )

def test_error_wrong_msg():
  try:
    th = TestHarnessSimple(
      Bits16, TestSrcCL, TestSinkCL,
      src_msgs  = [ b16(0xface), b16(0xface) ],
      sink_msgs = [ b16(0xface), b16(0xdead) ],
    )
    th.run_sim()
  except Exception as e:
    return
  raise Exception( 'Fail to detect error!' )

def test_error_late_msg():
  try:
    th = TestHarnessSimple(
      Bits16, TestSrcCL, TestSinkCL,
      src_msgs  = [ b16(0xface), b16(0xface) ],
      sink_msgs = [ b16(0xface), b16(0xdead) ],
    )
    th.set_param( 'top.src.construct', initial_delay=5 )
    th.set_param( 'top.sink.construct', arrival_time=[1,2] )
    th.run_sim()
  except Exception as e:
    return
  raise Exception( 'Fail to detect error!')

#-------------------------------------------------------------------------
# Customized compare function test
#-------------------------------------------------------------------------

def test_customized_cmp():
  th = TestHarnessSimple(
    Bits4, TestSrcCL, TestSinkCL,
    src_msgs  = [ b4(0b1110), b4(0b1111) ],
    sink_msgs = [ b4(0b0010), b4(0b0011) ],
  )
  th.set_param( 'top.sink.construct', cmp_fn=lambda a, b: a[0:2] == b[0:2] )
  th.run_sim()



#=========================================================================
# YosysStructuralTranslatorL2.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog structural translator."""

from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.translation.structural.SVStructuralTranslatorL2 import (
    SVStructuralTranslatorL2,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt

from .YosysStructuralTranslatorL1 import YosysStructuralTranslatorL1


class YosysStructuralTranslatorL2(
    YosysStructuralTranslatorL1, SVStructuralTranslatorL2 ):

  #-----------------------------------------------------------------------
  # Connection helper method
  #-----------------------------------------------------------------------

  def vec_conn_vector_gen( s, d, c_nbits, pid, wid, idx, dtype ):
    nbits = dtype.get_length()
    assert c_nbits - nbits >= 0
    msb, lsb = c_nbits-1, c_nbits-nbits
    return [ {
      "direction" : d,
      "pid" : pid,
      "wid" : wid,
      "idx" : idx + f"[{msb}:{lsb}]",
      "present" : True
    } ]

  def vec_conn_struct_gen( s, d, c_nbits, pid, wid, idx, dtype ):
    ret = []
    for name, field in dtype.get_all_properties().items():
      ret += s.vec_conn_dtype_gen( d, c_nbits, pid+"__"+name, wid, idx, field )
      c_nbits -= field.get_length()
    return ret

  def vec_conn_packed_gen( s, d, c_nbits, pid, wid, idx, _dtype ):

    def _packed_gen( d, c_nbits, pid, wid, idx, n_dim, p_nbits, dtype ):
      if not n_dim:
        return s.vec_conn_dtype_gen( d, c_nbits, pid, wid, idx, dtype )
      else:
        ret = []
        dec_nbits = p_nbits // n_dim[0]
        for i in reversed( range( n_dim[0]) ):
          ret += \
            _packed_gen(d, c_nbits, pid+"__"+str(i), wid, idx, n_dim[1:], dec_nbits, dtype)
          c_nbits -= dec_nbits
          assert c_nbits >= 0
        return ret

    p_n_dim = _dtype.get_dim_sizes()
    p_nbits = _dtype.get_length()
    dtype = _dtype.get_sub_dtype()
    return _packed_gen( d, c_nbits, pid, wid, idx, p_n_dim, p_nbits, dtype )

  def vec_conn_dtype_gen( s, d, c_nbits, pid, wid, idx, dtype ):
    if isinstance( dtype, rdt.Vector ):
      return s.vec_conn_vector_gen( d, c_nbits, pid, wid, idx, dtype )
    elif isinstance( dtype, rdt.Struct ):
      return s.vec_conn_struct_gen( d, c_nbits, pid, wid, idx, dtype )
    elif isinstance( dtype, rdt.PackedArray ):
      return s.vec_conn_packed_gen( d, c_nbits, pid, wid, idx, dtype )

    raise TypeError( f"unrecognized data type {dtype}!" )

  def struct_conn_gen( s, d, pid, wid, idx, dtype ):
    ret = []
    for name, field in dtype.get_all_properties().items():
      ret += s.dtype_conn_gen( d, pid+"__"+name, wid+"__"+name, idx, field )
    cur_nbits = dtype.get_length()
    for name, field in dtype.get_all_properties().items():
      ret += s.vec_conn_dtype_gen( d, cur_nbits, pid+"__"+name, wid, idx, field )
      cur_nbits -= field.get_length()
    assert cur_nbits == 0
    return ret

  def _packed_conn_gen( s, d, pid, wid, idx, n_dim, dtype ):
    if not n_dim:
      return s.dtype_conn_gen( d, pid, wid, idx, dtype )
    else:
      ret = []
      for i in range( n_dim[0] ):
        _pid = f"{pid}__{i}"
        _idx = f"{idx}[{i}]"
        ret += s._packed_conn_gen( d, _pid, wid, _idx, n_dim[1:], dtype )
      return ret

  def packed_conn_gen( s, d, pid, wid, idx, _dtype ):
    n_dim = _dtype.get_dim_sizes()
    dtype = _dtype.get_sub_dtype()
    return s._packed_conn_gen( d, pid, wid, idx, n_dim, dtype )

  def dtype_conn_gen( s, d, pid, wid, idx, dtype ):
    if isinstance( dtype, rdt.Struct ):
      return s.struct_conn_gen( d, pid, wid, idx, dtype )
    elif isinstance( dtype, rdt.PackedArray ):
      return s.packed_conn_gen( d, pid, wid, idx, dtype )
    else:
      return super().dtype_conn_gen( d, pid, wid, idx, dtype )

  #-----------------------------------------------------------------------
  # Port wire declaration helper method
  #-----------------------------------------------------------------------

  def wire_struct_gen( s, id_, dtype, n_dim ):
    ret = []
    # Generate wire for each field
    for name, field in dtype.get_all_properties().items():
      ret += s.wire_dtype_gen( id_+"__"+name, field, n_dim )
    # Generate a long vector for this struct signal
    ret.append( {
      "msb" : dtype.get_length()-1,
      "id_" : id_,
      "n_dim" : n_dim,
      "present" : True
    } )
    return ret

  def wire_packed_gen( s, id_, _dtype, n_dim ):
    _n_dim = _dtype.get_dim_sizes()
    dtype = _dtype.get_sub_dtype()
    ret = s.wire_dtype_gen( id_, dtype, n_dim + _n_dim )
    # Rigth now don't generate dedicated wire for the whole packed array -
    # we generate an unpacked array to group together all signals. Users
    # should access each element of the array instead of manipulating
    # this array as a whole.
    # ret.append( {
      # "msb" : _dtype.get_length()-1,
      # "id_" : id_,
      # "n_dim" : n_dim,
      # "present" : True
    # } )
    return ret

  def wire_dtype_gen( s, id_, dtype, n_dim ):
    if isinstance( dtype, rdt.Struct ):
      s.check_decl( id_, "" )
      return s.wire_struct_gen( id_, dtype, n_dim )
    elif isinstance( dtype, rdt.PackedArray ):
      s.check_decl( id_, "" )
      return s.wire_packed_gen( id_, dtype, n_dim )
    else:
      return \
        super().wire_dtype_gen(id_, dtype, n_dim )

  #-----------------------------------------------------------------------
  # Port declaration helper methods
  #-----------------------------------------------------------------------

  def _packed_gen( s, d, id_, n_dim, dtype ):
    if not n_dim:
      return s.dtype_gen( d, id_, dtype )
    else:
      ret = []
      for i in range(n_dim[0]):
        ret += s._packed_gen( d, f"{id_}__{i}", n_dim[1:], dtype )
      return ret

  def packed_gen( s, d, id_, _dtype ):
    n_dim = _dtype.get_dim_sizes()
    dtype = _dtype.get_sub_dtype()
    return s._packed_gen( d, id_, n_dim, dtype )

  def struct_gen( s, d, id_, dtype ):
    ret = []
    for name, field in dtype.get_all_properties().items():
      ret += s.dtype_gen( d, id_+"__"+name, field )
    return ret

  def dtype_gen( s, d, id_, dtype ):
    if isinstance( dtype, rdt.Struct ):
      s.check_decl( id_, "" )
      return s.struct_gen(d, id_, dtype)
    elif isinstance( dtype, rdt.PackedArray ):
      s.check_decl( id_, "" )
      return s.packed_gen(d, id_, dtype)
    else:
      return super().dtype_gen(d, id_, dtype)

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_packed_index( s, base_signal, index ):
    s.deq[-1]['s_index'] += "[{}]"
    s.deq[-1]['index'].append( int(index) )
    return f'{base_signal}[{index}]'

  def rtlir_tr_struct_attr( s, base_signal, attr ):
    s.deq[-1]['s_attr'] += "__{}"
    s.deq[-1]['attr'].append( attr )
    return f'{base_signal}.{attr}'

  def rtlir_tr_struct_instance( s, dtype, struct, first_called = True ):
    def _gen_packed_array( dtype, n_dim, array ):
      if not n_dim:
        if isinstance( dtype, rdt.Vector ):
          s_attr = s.rtlir_tr_literal_number( dtype.nbits, array, False )
          return {'attr':[], 'index':[], 's_attr':s_attr, 's_index':""}
        elif isinstance( dtype, rdt.Struct ):
          s_attr = s.rtlir_tr_struct_instance( dtype, array, False )
          return {'attr':[], 'index':[], 's_attr':s_attr, 's_index':""}
        else:
          assert False, f"unrecognized data type {dtype}!"
      else:
        ret = []
        for i in reversed( range( n_dim[0]) ):
          _ret = _gen_packed_array( dtype, n_dim[1:], array[i] )
          ret.append( _ret["s_attr"] )
        if n_dim[0] > 1:
          cat_str = "{{ " + ", ".join( ret ) + " }}"
        else:
          cat_str = ", ".join( ret )
        return {'attr':[], 'index':[], 's_attr':cat_str, 's_index':""}
    fields = []
    for name, Type in dtype.get_all_properties().items():
      field = getattr( struct, name )
      if isinstance( Type, rdt.Vector ):
        _field = s.rtlir_tr_literal_number( Type.nbits, field, False )
      elif isinstance( Type, rdt.Struct ):
        _field = s.rtlir_tr_struct_instance( Type, field, False )
      elif isinstance( Type, rdt.PackedArray ):
        n_dim = Type.get_dim_sizes()
        sub_dtype = Type.get_sub_dtype()
        _field = _gen_packed_array( sub_dtype, n_dim, field )['s_attr']
      else:
        assert False, f"unrecognized data type {Type}!"
      fields.append( _field )
    if len( fields ) == 1:
      struct_str = fields[0]
    else:
      struct_str = "{{ " + ", ".join( fields ) + " }}"
    if first_called:
      s.deq.append( {'attr':[], 'index':[], 's_attr':struct_str, 's_index':""} )
    return struct_str



#=========================================================================
# SVStructuralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 3 SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from ....testcases import (
    CaseConnectArrayBits32FooIfcComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectValRdyIfcComp,
)
from ..SVStructuralTranslatorL3 import SVStructuralTranslatorL3


def run_test( case, m ):
  m.elaborate()
  SVStructuralTranslatorL3.is_sverilog_reserved = lambda s, x: x in sverilog_reserved
  tr = SVStructuralTranslatorL3( m )
  tr.clear( m )
  tr.translate_structural( m )

  ifcs = tr.structural.decl_ifcs[m]
  conns = tr.structural.connections[m]

  check_eq( ifcs,  case.REF_IFC  )
  check_eq( conns, case.REF_CONN )

@pytest.mark.parametrize(
  'case', [
    CaseConnectValRdyIfcComp,
    CaseConnectArrayBits32FooIfcComp,
    CaseConnectArrayNestedIfcComp,
  ]
)
def test_sverilog_structural_L3( case ):
  run_test( case, case.DUT() )



#=========================================================================
# BehavioralRTLIRTypeCheckL2Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 29, 2019
"""Provide L2 behavioral RTLIR type check pass."""

from collections import OrderedDict

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRTypeCheckL1Pass import BehavioralRTLIRTypeCheckVisitorL1


class BehavioralRTLIRTypeCheckL2Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_tmpvars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()

    visitor = BehavioralRTLIRTypeCheckVisitorL2(
      m,
      m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed,
      m._pass_behavioral_rtlir_type_check.rtlir_tmpvars
    )

    for blk in m.get_update_block_order():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL2( BehavioralRTLIRTypeCheckVisitorL1 ):
  def __init__( s, component, freevars, accessed, tmpvars ):
    super().__init__(component, freevars, accessed)
    s.tmpvars = tmpvars
    s.BinOp_max_nbits = (bir.Add, bir.Sub, bir.Mult, bir.Div, bir.Mod, bir.Pow,
                         bir.BitAnd, bir.BitOr, bir.BitXor)
    s.BinOp_left_nbits = ( bir.ShiftLeft, bir.ShiftRightLogic )
    s.type_expect = {}
    lhs_types = ( rt.Port, rt.Wire, rt.NetWire, rt.NoneType )

    s.type_expect[ 'Assign' ] = {
      'target' : ( lhs_types, 'lhs of assignment must be signal/tmpvar!' ),
      'value' : ( rt.Signal, 'rhs of assignment should be signal/const!' )
    }
    s.type_expect[ 'BinOp' ] = {
      'left' : ( rt.Signal, 'lhs of binop should be signal/const!' ),
      'right' : ( rt.Signal, 'rhs of binop should be signal/const!' ),
    }
    s.type_expect[ 'UnaryOp' ] = {
      'operand' : ( rt.Signal, 'unary op only applies to signals and consts!' )
    }
    s.type_expect[ 'For' ] = {
      'start' : ( rt.Const, 'the start of a for-loop must be a constant expression!' ),
      'end':( rt.Const, 'the end of a for-loop must be a constant expression!' ),
      'step':( rt.Const, 'the step of a for-loop must be a constant expression!' )
    }
    s.type_expect[ 'If' ] = {
      'cond' : ( rt.Signal, 'the condition of if must be a signal!' )
    }
    s.type_expect[ 'IfExp' ] = {
      'cond' : ( rt.Signal, 'the condition of if-exp must be a signal!' ),
      'body' : ( rt.Signal, 'the body of if-exp must be a signal!' ),
      'orelse' : ( rt.Signal, 'the else branch of if-exp must be a signal!' )
    }

  def eval_const_binop( s, l, op, r ):
    """Evaluate ( l op r ) and return the result as an integer."""
    assert type( l ) == int or isinstance( l, pymtl3_datatype.Bits )
    assert type( r ) == int or isinstance( r, pymtl3_datatype.Bits )
    op_dict = {
      bir.And       : 'and', bir.Or    : 'or',
      bir.Add       : '+',   bir.Sub   : '-',  bir.Mult : '*',  bir.Div  : '/',
      bir.Mod       : '%',   bir.Pow   : '**',
      bir.ShiftLeft : '<<',  bir.ShiftRightLogic : '>>',
      bir.BitAnd    : '&',   bir.BitOr : '|',  bir.BitXor : '^',
    }
    _op = op_dict[ type( op ) ]
    return eval( f'l{_op}r' )

  def visit_Assign( s, node ):
    # RHS should have the same type as LHS
    rhs_type = node.value.Type
    lhs_type = node.target.Type

    if isinstance( node.target, bir.TmpVar ):
      tmpvar_id = (node.target.name, node.target.upblk_name)
      if lhs_type != rt.NoneType() and lhs_type.get_dtype() != rhs_type.get_dtype():
        raise PyMTLTypeError( s.blk, node.ast,
          f'conflicting type {rhs_type} for temporary variable {node.target.name}({lhs_type})!' )

      # Creating a temporaray variable
      # Reminder that a temporary variable is essentially a wire. So we use
      # rt.Wire here instead of rt.NetWire
      node.target.Type = rt.Wire( rhs_type.get_dtype() )
      s.tmpvars[ tmpvar_id ] = rt.Wire( rhs_type.get_dtype() )
      node.Type = None

    else:
      # non-temporary assignment is an L1 thing
      super().visit_Assign( node )

  def visit_If( s, node ):
    # Can the type of condition be cast into bool?
    dtype = node.cond.Type.get_dtype()
    if not rdt.Bool()( dtype ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the condition of "if" cannot be converted to bool!'
      )
    node.Type = None

  def visit_For( s, node ):
    try:
      step = node.step._value
      if step == 0:
        raise PyMTLTypeError( s.blk, node.ast,
          'the step of for-loop cannot be zero!' )
    except AttributeError:
      raise PyMTLTypeError( s.blk, node.ast,
        'the step of for-loop must be a constant!' )
    node.Type = None

  def visit_LoopVar( s, node ):
    node.Type = rt.Const( rdt.Vector( 32 ), None )

  def visit_TmpVar( s, node ):
    tmpvar_id = (node.name, node.upblk_name)
    if tmpvar_id not in s.tmpvars:
      # This tmpvar is being created. Later when it is used, its type can
      # be read from the tmpvar type environment.
      node.Type = rt.NoneType()

    else:
      node.Type = s.tmpvars[ tmpvar_id ]

  def visit_IfExp( s, node ):
    # Can the type of condition be cast into bool?
    if not rdt.Bool()( node.cond.Type.get_dtype() ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the condition of "if-exp" cannot be converted to bool!' )

    # body and orelse must have the same type
    # if node.body.Type != node.orelse.Type:
    if not node.body.Type.get_dtype()( node.orelse.Type.get_dtype() ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the body and orelse of "if-exp" must have the same type!' )
    node.Type = node.body.Type

  def visit_UnaryOp( s, node ):
    if isinstance( node.op, bir.Not ):
      dtype = node.operand.Type.get_dtype()
      if not rdt.Bool()( dtype ):
        raise PyMTLTypeError( s.blk, node.ast,
          'the operand of "Logic-not" cannot be cast to bool!' )
      if dtype.get_length() != 1:
        raise PyMTLTypeError( s.blk, node.ast,
          'the operand of "Logic-not" is not a single bit!' )
      node.Type = rt.NetWire( rdt.Bool() )
    else:
      node.Type = node.operand.Type

  def visit_BoolOp( s, node ):
    for value in node.values:
      if not isinstance(value.Type, rt.Signal) or not rdt.Bool()(value.Type.get_dtype()):
        raise PyMTLTypeError( s.blk, node.ast,
          f"{value} of {value.Type} cannot be cast into bool!")
    node.Type = rt.NetWire( rdt.Bool() )

  def visit_BinOp( s, node ):
    op = node.op
    l_type = node.left.Type.get_dtype()
    r_type = node.right.Type.get_dtype()
    if not( rdt.Vector(1)( l_type ) and rdt.Vector(1)( r_type ) ):
      raise PyMTLTypeError( s.blk, node.ast,
        f"both sides of {op.__class__.__name__} should be of vector type!" )

    if not isinstance( op, s.BinOp_left_nbits ) and l_type != r_type:
      raise PyMTLTypeError( s.blk, node.ast,
        f"LHS and RHS of {op.__class__.__name__} should have the same type ({l_type} vs {r_type})!" )

    l_nbits = l_type.get_length()
    r_nbits = r_type.get_length()

    # Enforcing Verilog bitwidth inference rules
    res_nbits = 0
    if isinstance( op, s.BinOp_max_nbits ):
      res_nbits = max( l_nbits, r_nbits )
    elif isinstance( op, s.BinOp_left_nbits ):
      res_nbits = l_nbits
    else:
      raise Exception( 'RTLIRTypeCheck internal error: unrecognized op!' )

    try:
      # Both sides are constant expressions
      l_val = node.left._value
      r_val = node.rigth._value
      node._value = s.eval_const_binop( l_val, op, r_val )
      node.Type = rt.Const( rdt.Vector( res_nbits ) )
    except AttributeError:
      # Both sides are constant but the value cannot be determined statically
      if isinstance(node.left.Type, rt.Const) and isinstance(node.right.Type, rt.Const):
        node.Type = rt.Const( rdt.Vector( res_nbits ), None )

      # Variable
      else:
        node.Type = rt.NetWire( rdt.Vector( res_nbits ) )

  def visit_Compare( s, node ):
    l_type = node.left.Type.get_dtype()
    r_type = node.right.Type.get_dtype()
    if l_type != r_type:
      raise PyMTLTypeError( s.blk, node.ast,
        f"LHS and RHS of {node.op.__class__.__name__} have different types ({l_type} vs {r_type})!" )
    node.Type = rt.NetWire( rdt.Bool() )



"""Expose behavioral RTLIR generation and type check passes.

PyMTL user should only interact with passes exposed here.
"""

from .BehavioralRTLIRGenL5Pass import BehavioralRTLIRGenL5Pass as BehavioralRTLIRGenPass
from .BehavioralRTLIRTypeCheckL5Pass import (
    BehavioralRTLIRTypeCheckL5Pass as BehavioralRTLIRTypeCheckPass,
)
from .BehavioralRTLIRVisualizationPass import BehavioralRTLIRVisualizationPass



#=========================================================================
# RTLIRTranslator_L1_cases_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 23, 2019
"""Test the RTLIR translator."""

import pytest

from pymtl3.passes.rtlir.util.test_utility import expected_failure, get_parameter

from ..behavioral.test.BehavioralTranslatorL1_test import test_generic_behavioral_L1
from ..errors import RTLIRTranslationError
from ..structural.test.StructuralTranslatorL1_test import test_generic_structural_L1
from ..testcases import (
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
)
from .TestRTLIRTranslator import TestRTLIRTranslator


def run_test( case, m ):
  if not m._dsl.constructed:
    m.elaborate()
  tr = TestRTLIRTranslator(m)
  tr.translate( m )
  src = tr.hierarchy.src
  assert src == case.REF_SRC

@pytest.mark.parametrize(
  'case', get_parameter('case', test_generic_behavioral_L1) + \
          get_parameter('case', test_generic_structural_L1)
)
def test_generic_L1( case ):
  run_test( case, case.DUT() )

def test_bit_sel_over_bit_sel():
  with expected_failure( RTLIRTranslationError, "over bit/part selection" ):
    run_test( CaseBitSelOverBitSelComp, CaseBitSelOverBitSelComp.DUT() )

def test_bit_sel_over_part_sel():
  with expected_failure( RTLIRTranslationError, "over bit/part selection" ):
    run_test( CaseBitSelOverPartSelComp, CaseBitSelOverPartSelComp.DUT() )

def test_part_sel_over_bit_sel():
  with expected_failure( RTLIRTranslationError, "over bit/part selection" ):
    run_test( CasePartSelOverBitSelComp, CasePartSelOverBitSelComp.DUT() )

def test_part_sel_over_part_sel():
  with expected_failure( RTLIRTranslationError, "over bit/part selection" ):
    run_test( CasePartSelOverPartSelComp, CasePartSelOverPartSelComp.DUT() )



"""
========================================================================
AddSimUtilFuncsPass.py
========================================================================

Author : Shunning Jiang
Date   : Jan 26, 2020
"""

from pymtl3.datatypes import b1
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError


class AddSimUtilFuncsPass( BasePass ):
  def __init__( self, active_high=True ):
    assert active_high in [ True, False ]
    self.active_high = active_high

  def __call__( self, top ):
    if not hasattr( top, "tick" ):
      raise PassOrderError( "tick" )

    if hasattr(top, "sim_reset"):
      raise AttributeError( "Please modify the attribute top.sim_reset to "
                            "a different name.")
    if hasattr(top, "print_line_trace"):
      raise AttributeError( "Please modify the attribute top.print_line_trace to "
                            "a different name.")

    top.sim_reset = self.create_reset( top, self.active_high )
    top.print_line_trace = self.create_print_line_trace( top )

  @staticmethod
  # Simulation related APIs
  def create_reset( top, active_high ):
    def reset( print_line_trace=False ):
      if print_line_trace:
        print()
      top.reset = b1( active_high )
      top.tick() # Tick twice to propagate the reset signal
      if print_line_trace:
        print( f"{top.simulated_cycles:3}r {top.line_trace()}" )
      top.tick()
      top.reset = b1( not active_high )
    return reset

  @staticmethod
  # Simulation related APIs
  def create_print_line_trace( top ):
    def print_line_trace():
      print( f"{top.simulated_cycles:3}: {top.line_trace()}" )
    return print_line_trace



#=========================================================================
# CollectSignalPass_test.py
#=========================================================================
# A simple demo of how collect signal pass is used in the simulation of
# a toy component.
#
# Author : Kaishuo Cheng
# Date   : Oct 8th, 2019

from pymtl3.datatypes import Bits32, b32
from pymtl3.dsl import *
from pymtl3.passes import SimulationPass, TracingConfigs
from pymtl3.passes.PassGroups import SimulationPass


class Toy( Component ):
  """Toy adder component"""

  def construct( s ):
    # Interfaces
    s.in0 = InPort( Bits32 )
    s.in1 = InPort( Bits32 )
    s.out = OutPort( Bits32 )

    @s.update
    def add_upblk():
      # This update block models the behavior of a 32-bit adder
      s.out = s.in0 + s.in1

def process_binary(sig):
  """
  Returns int value from a signal in 32b form. Used for testing.
  Example: input: 0b00000000000000000000000000000000
           output: 0
  """
  tempint = int(sig[2:],2)
  if sig[2] == '1':
     #taking 2's complement.
     #leading 1 indicates a negative number
    return tempint -2 **32

  else:
    return tempint

def test_toy():
  # Create a toy component and elaborate it
  dut = Toy()

  # Turn on textwave
  dut.config_tracing = TracingConfigs( tracing='text_fancy' )

  dut.elaborate()

  # Setup the simulation
  dut.apply( SimulationPass() )
  dut.sim_reset()

  # Test vector
  vector = [
    #  in0        in1       out
    b32(1),    b32(2),    b32(3),
    b32(-1),   b32(2),    b32(1),
    b32(42),   b32(2),    b32(44),
    b32(1),    b32(-2),   b32(-1),
    b32(42),   b32(-42),  b32(0),
  ]

  testlist = [
  #in0
  [0,0,1,-1,42,1,42], #dut's reset is on for the first two cycles".

  #in1
  [0,0,2,2,2,-2,-42],

  #out
  [0,0,3,1,44,-1,0]
  ]
  # Begin simulation
  for in0, in1, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.in0 = in0
    dut.in1 = in1
    dut.eval_combinational()
    dut.tick()
    assert dut.out == out

  #test
  sig = dut._tracing.text_sigs
  siglist = ["s.in0","s.in1","s.out","s.clk","s.reset"]
  for i in siglist:
    assert i in sig,"signals not captured"

  print(sig)
  for i in range(3): # in0, in1, and out
    partsig = sig[ siglist[i] ]
    signal_length = len(partsig)
    assert signal_length >= 5,"missing some cycles of signals"

    for j in range(signal_length):
      assert testlist[i][j] == process_binary(partsig[j]),"collected wrong signals"

  print("All signals captured in top._tracing.text_sigs!")



"""
========================================================================
ComponentLevel5_test.py
========================================================================

Author : Shunning Jiang
Date   : Jan 4, 2019
"""
from collections import deque

from pymtl3.dsl.ComponentLevel3 import connect
from pymtl3.dsl.ComponentLevel5 import ComponentLevel5, method_port
from pymtl3.dsl.Connectable import CalleePort, CallerPort, Interface
from pymtl3.dsl.ConstraintTypes import M, U

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  print()
  T, time = 0, 20
  while not A.done() and T < time:
    A.tick()
    print(A.line_trace())
    T += 1
  return T

class SimpleTestSource( ComponentLevel5 ):

  def construct( s, msgs ):
    s.msgs = deque( msgs )

    s.req     = CallerPort()
    s.req_rdy = CallerPort()

    s.v = 0
    @s.update
    def up_src():
      s.v = None
      if s.req_rdy() and s.msgs:
        s.v = s.msgs.popleft()
        s.req( s.v )

  def done( s ):
    return not s.msgs

  def line_trace( s ):
    return "{:4}".format( "" if s.v is None else s.v )

class TestSinkError( Exception ):
  pass

class SimpleTestSink( ComponentLevel5 ):

  def resp_( s, msg ):
    ref = s.msgs[ s.idx ]
    s.idx += 1

    if msg != ref:
      raise TestSinkError( """
The test sink received an incorrect message!
- sink name    : {}
- msg number   : {}
- expected msg : {}
- actual msg   : {}
""".format( s, s.idx, ref, msg )
      )

  def resp_rdy_( s ):
    return True

  def construct( s, msgs ):
    s.msgs = list( msgs )
    s.idx  = 0

    s.resp     = CalleePort( method = s.resp_     )
    s.resp_rdy = CalleePort( method = s.resp_rdy_ )

  def done( s ):
    return s.idx >= len(s.msgs)

  def line_trace( s ):
    return ""

def test_simple_src_dumb_sink():

  class Top( ComponentLevel5 ):

    def construct( s ):
      s.src  = SimpleTestSource( [1,2,3,4] )
      s.sink = SimpleTestSink( [1,2,3,4] )

      s.src.req //= s.sink.resp
      connect( s.src.req_rdy, s.sink.resp_rdy )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return  s.src.line_trace() + " >>> " + s.sink.line_trace()


  assert _test_model( Top ) == 4 # regression: 4 cycles

class TestSinkUp( ComponentLevel5 ):

  def resp_rdy_( s ):
    return s.idx < len(s.msgs)

  def resp_( s, v ):
    s.queue.appendleft(v)

  def construct( s, msgs ):
    s.msgs  = list( msgs )
    s.queue = deque( maxlen=1 )
    s.idx  = 0

    s.resp     = CalleePort( method = s.resp_ )
    s.resp_rdy = CalleePort( method = s.resp_rdy_ )

    s.v = None

    @s.update
    def up_sink():
      s.v = None

      if s.queue:
        msg = s.queue.popleft()
        s.v = msg

        if s.idx >= len(s.msgs):
          raise TestSinkError( """
  The test sink received a message that !
  - sink name    : {}
  - msg number   : {}
  - actual msg   : {}
  """.format( s, s.idx, msg )
          )
        else:
          ref = s.msgs[ s.idx ]
          s.idx += 1

          if msg != ref:
            raise TestSinkError( """
  The test sink received an incorrect message!
  - sink name    : {}
  - msg number   : {}
  - expected msg : {}
  - actual msg   : {}
  """.format( s, s.idx, ref, msg )
          )

    s.add_constraints(
      U(up_sink) < M(s.resp_    ), # pipe behavior
      U(up_sink) < M(s.resp_rdy_),
    )

  def done( s ):
    return s.idx >= len(s.msgs)

  def line_trace( s ):
    return "{:4}".format( "" if s.v is None else s.v )

def test_simple_src_up_sink():

  class Top( ComponentLevel5 ):

    def construct( s ):
      s.src  = SimpleTestSource( [1,2,3,4] )
      s.sink = TestSinkUp( [1,2,3,4] )

      connect( s.src.req,     s.sink.resp )
      connect( s.src.req_rdy, s.sink.resp_rdy )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return  s.src.line_trace() + " >>> " + s.sink.line_trace()

  assert _test_model( Top ) == 5 # regression: 5 cycles

class PassThrough( ComponentLevel5 ):

  @method_port
  def req( s, msg ):
    assert s.resp_rdy()
    s.resp( msg )

  @method_port
  def req_rdy( s ):
    return s.resp_rdy()

  def construct( s ):
    s.resp     = CalleePort()
    s.resp_rdy = CalleePort()
    s.entry = None

    s.add_constraints(
      M(s.req) == M(s.resp),
      M(s.req_rdy) == M(s.resp_rdy),
    )

def test_constraint_equal_pass_through():

  class Top( ComponentLevel5 ):

    def construct( s ):
      s.src  = SimpleTestSource( [1,2,3,4] )
      s.mid  = PassThrough()( req = s.src.req, req_rdy = s.src.req_rdy )
      s.sink = TestSinkUp( [1,2,3,4] )( resp = s.mid.resp,
                                        resp_rdy = s.mid.resp_rdy )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return  s.src.line_trace() + " >>> " + s.sink.line_trace()

  assert _test_model( Top ) == 5 # regression: 5 cycles

def test_constraint_equal_pass_way_through():

  class Top( ComponentLevel5 ):

    def construct( s ):
      s.src  = SimpleTestSource( [1,2,3,4] )
      s.mid  = [ PassThrough() for _ in range(5) ]
      s.sink = TestSinkUp( [1,2,3,4] )


      connect( s.src.req,     s.mid[0].req )
      connect( s.src.req_rdy, s.mid[0].req_rdy )

      for i in range(4):
        s.mid[i].resp     //= s.mid[i+1].req
        s.mid[i].resp_rdy //= s.mid[i+1].resp_rdy

      s.mid[4].resp     //=  s.sink.resp
      s.mid[4].resp_rdy //= s.sink.resp_rdy

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return  s.src.line_trace() + " >>> " + s.sink.line_trace()

  assert _test_model( Top ) == 5 # regression: 5 cycles

def test_method_interface():

  class RecvIfcCL( Interface ):
    def construct( s, recv=None, rdy=None):
      s.recv = CalleePort( method = recv )
      s.rdy  = CalleePort( method = rdy )

    # Here we customize method interface connections
    def connect( s, other, parent ):
      if isinstance( other, SendIfcCL ):
        connect( s.recv, other.send )
        connect( s.rdy,  other.rdy )
        return True

      return False

  class SendIfcCL( Interface ):
    def construct( s ):
      s.send = CallerPort()
      s.rdy  = CallerPort()

    def connect( s, other, parent ):
      if isinstance( other, RecvIfcCL ):
        connect( s.send, other.recv )
        connect( s.rdy,  other.rdy )
        return True

      return False

  class SimpleTestSourceIfc( ComponentLevel5 ):

    def construct( s, msgs ):
      s.msgs = deque( msgs )

      s.req = SendIfcCL()

      s.v = 0
      @s.update
      def up_src():
        s.v = None
        if s.req.rdy() and s.msgs:
          s.v = s.msgs.popleft()
          s.req.send( s.v )

    def done( s ):
      return not s.msgs

    def line_trace( s ):
      return "{:4}".format( "" if s.v is None else s.v )

  class SimpleTestSinkIfc( ComponentLevel5 ):

    def resp_( s, msg ):
      ref = s.msgs[ s.idx ]
      s.idx += 1

      if msg != ref:
        raise TestSinkError( """
  The test sink received an incorrect message!
  - sink name    : {}
  - msg number   : {}
  - expected msg : {}
  - actual msg   : {}
  """.format( s, s.idx, ref, msg )
        )

    def resp_rdy_( s ):
      return True

    def construct( s, msgs ):
      s.msgs = list( msgs )
      s.idx  = 0

      s.resp = RecvIfcCL( recv = s.resp_, rdy = s.resp_rdy_ )

    def done( s ):
      return s.idx >= len(s.msgs)

    def line_trace( s ):
      return ""

  class Top( ComponentLevel5 ):

    def construct( s ):
      s.src  = SimpleTestSourceIfc( [1,2,3,4] )
      s.sink = SimpleTestSinkIfc( [1,2,3,4] )

      connect( s.src.req, s.sink.resp )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return  s.src.line_trace() + " >>> " + s.sink.line_trace()

  assert _test_model( Top ) == 4 # regression: 4 cycles



#=========================================================================
# StructuralRTLIRGenL1Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 19, 2019
"""Test the generation of level 1 structural RTLIR."""

from collections import defaultdict, deque

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits1, Bits4, Bits32
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt
from pymtl3.passes.rtlir.structural import StructuralRTLIRSignalExpr as sexp
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL1Pass import (
    StructuralRTLIRGenL1Pass,
)
from pymtl3.passes.testcases import (
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectPortIndexComp,
    CaseConnectSliceToOutComp,
    CaseConstBits32AttrComp,
    CaseInx2Outx2ConnectComp,
)


def gen_connections( top ):
  _inst_conns = defaultdict( set )

  # Generate the connections assuming no sub-components
  nets = top.get_all_value_nets()
  adjs = top.get_signal_adjacency_dict()

  for writer, net in nets:
    S = deque( [ writer ] )
    visited = {  writer  }
    while S:
      u = S.pop()
      writer_host        = u.get_host_component()
      writer_host_parent = writer_host.get_parent_object()
      for v in adjs[u]:
        if v not in visited:
          visited.add( v )
          S.append( v )
          reader_host        = v.get_host_component()
          reader_host_parent = reader_host.get_parent_object()

          # Four possible cases for the reader and writer signals:
          # 1.   They have the same host component. Both need
          #       to be added to the host component.
          # 2/3. One's host component is the parent of the other.
          #       Both need to be added to the parent component.
          # 4.   They have the same parent component.
          #       Both need to be added to the parent component.

          if writer_host is reader_host:
            _inst_conns[writer_host].add( ( u, v ) )
          elif writer_host_parent is reader_host:
            _inst_conns[reader_host].add( ( u, v ) )
          elif writer_host is reader_host_parent:
            _inst_conns[writer_host].add( ( u, v ) )
          elif writer_host_parent == reader_host_parent:
            _inst_conns[writer_host_parent].add( ( u, v ) )
          else:
            raise TypeError( "unexpected connection type!" )

  return _inst_conns

def test_L1_const_numbers():
  a = CaseConstBits32AttrComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  assert ns.consts == [('const', rt.Array([5], rt.Const(rdt.Vector(32))), a.const)]

def test_L1_connection_order():
  a = CaseInx2Outx2ConnectComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  assert ns.connections == \
    [(sexp.CurCompAttr(comp, 'in_1'), sexp.CurCompAttr(comp, 'out1')),
     (sexp.CurCompAttr(comp, 'in_2'), sexp.CurCompAttr(comp, 'out2'))]

def test_L1_port_index():
  a = CaseConnectPortIndexComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  assert ns.connections == \
    [(sexp.PortIndex(sexp.CurCompAttr(comp, 'in_'), 2), sexp.CurCompAttr(comp, 'out'))]

def test_L1_wire_index():
  a = CaseConnectInToWireComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  assert ns.connections[0] == \
    (sexp.WireIndex(sexp.CurCompAttr(comp, 'wire_'), 2), sexp.CurCompAttr(comp, 'out'))

def test_L1_const_index():
  a = CaseConnectConstToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  # The expression structure is removed and only the constant value
  # is left in this node.
  assert ns.connections == \
    [(sexp.ConstInstance(a.const_[2], 42), sexp.CurCompAttr(comp, 'out'))]

def test_L1_bit_selection():
  a = CaseConnectBitSelToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  # PyMTL DSL converts bit selection into 1-bit part selection!
  assert ns.connections == \
    [(sexp.PartSelection(sexp.CurCompAttr(comp, 'in_'), 0, 1), sexp.CurCompAttr(comp, 'out'))]

def test_L1_part_selection():
  a = CaseConnectSliceToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  assert ns.connections == \
    [(sexp.PartSelection(sexp.CurCompAttr(comp, 'in_'), 4, 8), sexp.CurCompAttr(comp, 'out'))]

def test_L1_bits_connection():
  a = CaseConnectBitsConstToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL1Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = sexp.CurComp(a, 's')
  assert ns.connections == \
    [(sexp.ConstInstance(Bits32(0), 0), sexp.CurCompAttr(comp, 'out'))]



#=========================================================================
# StructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 3 structural translators."""

import pytest

from ...testcases import (
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectSubCompIfcHierarchyComp,
)
from ..StructuralTranslatorL4 import StructuralTranslatorL4
from .TestStructuralTranslator import mk_TestStructuralTranslator


def run_test( case, m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL4)(m)
  tr.clear( m )
  tr.translate_structural(m)
  connections = tr.structural.connections[m]
  assert connections == case.REF_CONN
  decl_comp = tr.structural.decl_subcomps[m]
  assert decl_comp == case.REF_COMP

@pytest.mark.parametrize(
  'case', [
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectSubCompIfcHierarchyComp,
  ]
)
def test_generic_structural_L4( case ):
  run_test( case, case.DUT() )



"""
========================================================================
Slicing_test.py
========================================================================

Author : Shunning Jiang
Date   : Aug 23, 2018
"""
from pymtl3.datatypes import Bits2, Bits4, Bits14, Bits16, Bits24, Bits32
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import Wire
from pymtl3.dsl.errors import MultiWriterError, NoWriterError

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  for i in range(10):
    A.tick()

# write two disjoint slices
def test_write_two_disjoint_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_16():
        s.A[0:16] = Bits16( 0xff )

      @s.update
      def up_wr_16_30():
        s.A[16:30] = Bits14( 0xff )

      @s.update
      def up_rd_12_30():
        assert s.A[12:30] == 0xff0

  _test_model( Top )

# write two disjoint slices, but one slice is not read at all
def test_write_two_disjoint_slices_no_reader():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_16():
        s.A[0:16] = Bits16( 0xff )

      @s.update
      def up_wr_16_30():
        s.A[16:30] = Bits14( 0xff )

      @s.update
      def up_rd_17_30():
        assert s.A[16:30] == 0xff

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_16_30" and \
          # m._all_id_upblk[y].__name__ == "up_rd_17_30" # only one constraint


# write two overlapping slices
def test_write_two_overlapping_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_24():
        s.A[0:24] = Bits24( 0xff )

      @s.update
      def up_wr_8_32():
        s.A[8:32] = Bits24( 0xff )

      @s.update
      def up_rd_A():
        x = s.A

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# write two slices and a single bit
def test_write_two_slices_and_bit():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_16():
        s.A[0:16] = Bits16( 0xff )

      @s.update
      def up_wr_16_30():
        s.A[16:30] = Bits14( 0xff )

      @s.update
      def up_wr_30_31():
        s.A[30] = Bits1( 1 )

      @s.update
      def up_rd_A():
        print(s.A[0:17])

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 2
  # _, x = list(m._all_constraints)[0]
  # _, y = list(m._all_constraints)[1]

  # two constraints are: up_wr_0_16 < up_rd_A and up_wr_16_30 < up_rd_A
  # assert  m._all_id_upblk[x].__name__ == "up_rd_A" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A"

# write a slice and a single bit, but they are overlapped
def test_write_slices_and_bit_overlapped():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_16():
        s.A[0:16] = Bits16( 0xff )

      @s.update
      def up_wr_15():
        s.A[15] = Bits1( 1 )

      @s.update
      def up_rd_A():
        print(s.A[0:17])

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# write a slice and there are two reader
def test_multiple_readers():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_8_24():
        s.A[8:24] = Bits16( 0x1234 )

      @s.update
      def up_rd_0_12():
        assert s.A[0:12] == 0x400

      @s.update
      def up_rd_bunch():
        assert s.A[23] == 0
        assert s.A[22] == 0
        assert s.A[21] == 0
        assert s.A[20] == 1
        assert s.A[19] == 0
        assert s.A[18] == 0
        assert s.A[17] == 1
        assert s.A[16] == 0

  _test_model( Top )

# 1. WR A[s], RD A    (A[s] (=) A, SAME AS data struct)
#    WR A[s], WR A    (detect 2-writer conflict, SAME AS data struct)
#    WR A[s], RD A[t] (A[s] (=) A[t] if s intersects t)
#    WR A[s], WR A[t] (detect 2-writer conflict if s intersects t)

# 2. WR A   , RD A[s] (A[s] (=) A, SAME AS data struct)

# 3. WR A[s], A   |=y, RD y (mark A as writer in net {A,y}, SAME AS data struct)
#    WR A[s], A   |=y, WR y (detect 2-writer conflict, SAME AS data struct)
#    WR A[s], A[t]|=y, RD y (mark A[t] as writer in net {A[t],y} if s intersects t)
#    WR A[s], A[t]|=y, WR y (detect 2-writer conflict if s intersects t)

# 4. WR A   , A[s]|=y, RD y (mark A[s] as writer in net {A[s],y}, SAME AS data struct)
#    WR A   , A[s]|=y, WR y (detect 2-writer conflict, SAME AS data struct)

# 5. WR x, x|=A[s], RD A    (A[s] (=) A, SAME AS data struct)
#    WR x, x|=A[s], RD A[t] (A[s] (=) A[t] if s intersects t)

# 6. WR x, x|=A   , RD A[s] (A[s] (=) A, SAME AS data struct)

# 7. WR x, x|=A[s], A   |=y, RD y (mark A as writer and implicit constraint)
#    WR x, x|=A[s], A   |=y, WR y (detect 2-writer conflict)
#    WR x, x|=A[s], A[t]|=y, RD y (mark A[t] as writer and implicit constraint if s intersects t)
#    WR x, x|=A[s], A[t]|=y, WR y (detect 2-writer conflict if s intersects t)

# 8. WR x, x|=A   , A[s]|=y, RD y (mark A[s] as writer in net {A[s],y}, SAME AS data struct)

# --------------------------------------------------------------------------

# RD A[s]
#  - WR A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A[t]       (A[s] (=) A[t]                          if s intersects t)
#  - A   |=x, WR x (A[s] (=) A,                          SAME AS data struct)
#  - A[t]|=x, WR x (A[s] (=) A[t]                          if s intersects t)

# WR A[s]
#  - RD A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A          (detect 2-writer conflict,            SAME AS data struct)
#  - WR A[t]       (detect 2-writer conflict               if s intersects t)
#  - A   |=x       (mark A as writer in net {A,x},       SAME AS data struct)
#  - A   |=x, WR x (detect 2-writer conflict,            SAME AS data struct)
#  - A[t]|=x       (mark A[t] as writer in net {A[t],x}    if s intersects t)
#  - A[t]|=x, WR x (detect 2-writer conflict               if s intersects t)

# A[s]|=x
#  - WR A          (mark A[s] as writer in net {A[s],x}, SAME AS data struct)
#  - A|=y, WR y    (mark A[s] as writer in net {A[s],x}, SAME AS data struct)
#  - A[t]|=y, WR y (mark A[s] as writer in net {A[s],x},   if s intersects t)

# A[s]|=x, WR x
#  - RD A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A          (detect 2-writer conflict,            SAME AS data struct)
#  - A   |=y       (mark A as writer in net {A,y}        SAME AS data struct)
#  - A   |=y, WR y (detect 2-writer conflict,            SAME AS data struct)
#  - A[t]|=y, WR y (detect 2-writer conflict               if s intersects t)

# RD A[s] - WR A
def test_rd_As_wr_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_A():
        s.A = Bits32( 123 )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 123

  _test_model( Top )

# RD A[s] - WR A[t], intersect
def test_rd_As_wr_At_impl_intersect():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_At():
        s.A[8:24] = Bits16( 0xff )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 0xff00

  _test_model( Top )

# RD A[s] - WR A[t], not intersect
def test_rd_As_wr_At_impl_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_At():
        s.A[16:32] = Bits16( 0xff )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 0 # no constraint at all!

# WR A[s] - WR A
def test_wr_As_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_wr_A():
        s.A = Bits32( 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - WR A[t], intersect
def test_wr_As_wr_At_intersect():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_wr_At():
        s.A[2:4] = Bits2( 2 )

      @s.update
      def up_rd_A():
        z = s.A

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - WR A[t], not intersect
def test_wr_As_wr_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_wr_At():
        s.A[5:7] = Bits2( 2 )

      @s.update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# WR A[s] - RD A
def test_wr_As_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# WR A[s] - RD A, RD A[t], intersect
def test_wr_As_rd_A_rd_At_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_rd_A():
        z = s.A

      @s.update
      def up_rd_As():
        assert s.A[2:4] == 1

  _test_model( Top )

# WR A[s] - RD A, RD A[t], not intersect
def test_wr_As_rd_A_rd_At_cannot_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[1:3] = Bits2( 2 )

      @s.update
      def up_rd_A():
        z = s.A

      @s.update
      def up_rd_At():
        assert s.A[3:5] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_As" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A" # only one constraint

# WR A - RD A[s], RD A[t]
def test_wr_A_rd_slices_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_A():
        s.A = Bits32( 0x12345678 )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 0x5678

      @s.update
      def up_rd_At():
        assert s.A[8:24] == 0x3456

  _test_model( Top )

# WR A[s] - RD A, RD A[t], not intersect
def test_wr_As_rd_A_rd_At_bit_cannot_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_As():
        s.A[0:16] = Bits16( 0x1234 )

      @s.update
      def up_rd_A():
        z = s.A

      @s.update
      def up_rd_At():
        assert s.A[16] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_As" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A" # only one constraint

# RD A[s] - A|=x, WR x
def test_connect_rd_As_wr_x_conn_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.A, s.x )

      @s.update
      def up_wr_x():
        s.x = Bits32( 123 )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 123

  _test_model( Top )

# RD A[s] - A[t]|=x, WR x, intersect
def test_connect_rd_As_wr_x_conn_At_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[0:24], s.x )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_rd_As():
        assert s.A[0:16] == 0x3456

  _test_model( Top )

# RD A[s] - A[t]|=x, WR x, not intersect
def test_connect_rd_As_wr_x_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[0:24], s.x )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_rd_As():
        assert s.A[24:32] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_x" and \
          # m._all_id_upblk[y].__name__ == "s_x__1" # connection block

# WR A[s] - A|=x
def test_connect_wr_As_rd_x_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A )

      @s.update
      def up_wr_As():
        s.A[0:24] = Bits24( 0x123456 )

  _test_model( Top )

# WR A[s] - A|=x, WR x
def test_connect_wr_As_wr_x_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A )

      @s.update
      def up_wr_As():
        s.A[0:24] = Bits24( 0x123456 )

      @s.update
      def up_wr_x():
        s.x = Bits32( 0x87654321 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - A[t]|=x, intersect
def test_connect_wr_As_rd_x_conn_At_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @s.update
      def up_wr_As():
        s.A[0:24] = Bits24( 0x123456 )

  _test_model( Top )

# WR A[s] - A[t]|=x, not intersect
def test_connect_wr_As_rd_x_conn_At_no_driver():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @s.update
      def up_wr_As():
        s.A[0:4] = Bits4( 0xf )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

# WR A[s] - A[t]|=x, WR x, intersect
def test_connect_wr_As_wr_x_conn_At_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @s.update
      def up_wr_As():
        s.A[0:24] = Bits24( 0x123456 )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x654321 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - A[t]|=x, WR x, not intersect
def test_connect_wr_As_wr_x_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @s.update
      def up_wr_As():
        s.A[0:4] = Bits4( 0xf )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x654321 )

      @s.update
      def up_rd_A():
        assert s.A == 0x6543210f

  _test_model( Top )

# A[s]|=x, WR x - RD A
def test_connect_wr_x_conn_As_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_rd_A():
        assert s.A == 0x12345600

  _test_model( Top )

# A[s]|=x, WR x - WR A
def test_connect_wr_x_conn_As_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.y  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.x, s.y )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_wr_A():
        s.A = Bits32( 0x12345678 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x - WR A
def test_connect_rd_x_conn_As_wr_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )

      @s.update
      def up_wr_A():
        s.A = Bits32( 0x12345678 )

      @s.update
      def up_rd_x():
        assert s.x == 0x123456

  _test_model( Top )

# A[s]|=x, WR x - A|=y, WR y
def test_connect_wr_x_conn_As_wr_y_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A      , s.y )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_wr_y():
        s.y = Bits32( 0x12345678 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x, WR x - A[t]|=y, WR y, intersect
def test_connect_wr_x_conn_As_wr_y_conn_At_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits16 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:16], s.y )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_wr_y():
        s.y = Bits16( 0x1234 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x, WR x - A[t]|=y, WR y, not intersect
def test_connect_wr_x_conn_As_wr_y_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits4 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:4],  s.y )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_wr_y():
        s.y = Bits4( 0xf )

      @s.update
      def up_rd_A():
        assert s.A == 0x1234560f

  _test_model( Top )

# A[s]|=x, WR x - A|=y, RD y
def test_connect_wr_x_conn_As_rd_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A,       s.y )

      @s.update
      def up_wr_x():
        s.x = Bits24( 0x123456 )

      @s.update
      def up_rd_y():
        assert s.y == 0x12345600

  _test_model( Top )

# A[s]|=x - A|=y, WR y
def test_connect_rd_x_conn_As_wr_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A,       s.y )

      @s.update
      def up_rd_x():
        assert s.x == 0x123456

      @s.update
      def up_wr_y():
        s.y = Bits32( 0x12345678 )

  _test_model( Top )

# A[s]|=x - A[t]|=y, WR y, intersect
def test_connect_rd_x_conn_As_wr_y_conn_At_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits16 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:16], s.y )

      @s.update
      def up_rd_x():
        assert s.x == 0x12

      @s.update
      def up_wr_y():
        s.y = Bits16( 0x1234 )

  _test_model( Top )

# A[s]|=x - A[t]|=y, WR y, not intersect
def test_connect_rd_x_conn_As_wr_y_conn_no_driver():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits4 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:4 ], s.y )

      @s.update
      def up_rd_x():
        assert s.x == 0

      @s.update
      def up_wr_y():
        s.y = Bits4( 0xf )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

def test_iterative_find_nets():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.w  = Wire( Bits32 )
      s.x  = Wire( Bits32 )
      s.y  = Wire( Bits32 )
      s.z  = Wire( Bits32 )

      connect( s.w[0:16],  s.x[8:24] ) # net1
      connect( s.x[16:32], s.y[0:16] ) # net2
      connect( s.y[8:24],  s.z[0:16] ) # net3

      @s.update
      def up_wr_s_w():
        s.w = Bits32( 0x12345678 )

  _test_model( Top )

def test_multiple_sibling_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.A  = Wire( Bits32 )
      s.x  = Wire( Bits16 )
      s.y  = Wire( Bits16 )
      s.z  = Wire( Bits16 )

      connect( s.A[0:16], s.x ) # net1

      connect( s.A[8:24], s.y ) # net2

      connect( s.A[16:32], s.z ) # net3

      @s.update
      def up_wr_s_w():
        s.x = Bits16( 0x1234 )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

def test_multiple_write_same_slice():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = Wire( Bits32 )
      s.wire = Wire( Bits32 )
      connect( s.out, s.wire[0:32] )

      @s.update
      def comb_upblk():
        s.wire[0:16]  = 0
        s.wire[16:32] = 1

  a = A()
  a.elaborate()

def test_multiple_write_same_slice_with_overlap():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = Wire( Bits32 )
      s.wire = Wire( Bits32 )
      connect( s.out, s.wire[0:32] )

      s.wire2 = Wire( Bits24 )
      connect( s.wire2[0:24], s.wire[8:32] )

      @s.update
      def comb_upblk():
        s.wire[0:16]  = 0
        s.wire2[0:24] = 1

  try:
    _test_model( A )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")
